/* #undef SEACAS_HAVE_MPI */

/* #undef IOSS_THREADSAFE */

/* Define to indicate that the SEACAS project is being built with the KOKKOS package enabled*/
/* #undef SEACAS_HAVE_KOKKOS */

/* #undef SEACAS_HAVE_FAODEL */

/* #undef SEACAS_HAVE_DATAWARP */

#define SEACAS_HAVE_EXODUS

/* #undef SEACAS_HAVE_EXONULL */

#define SEACAS_HAVE_CGNS

/* #undef SEACAS_HAVE_PAMGEN */

/* #undef HAVE_SEACASIOSS_ADIOS2 */

/* #undef SEACAS_HAVE_CATALYST2 */
