// SPDX-FileCopyrightText: Copyright (c) Ken Martin, Will Schroeder, Bill Lorensen
// SPDX-License-Identifier: BSD-3-Clause
#ifndef vtkTextureObjectVS_h
#define vtkTextureObjectVS_h

#include "vtkRenderingOpenGL2Module.h"
#include "vtkABINamespace.h"

VTK_ABI_NAMESPACE_BEGIN

VTKRENDERINGOPENGL2_EXPORT extern const char *vtkTextureObjectVS;

VTK_ABI_NAMESPACE_END

#endif
