#ifndef vtk_egl_mangle_h
#define vtk_egl_mangle_h

#define GLAD_EGL_VERSION_1_0 VTK_GLAD_EGL_VERSION_1_0
#define GLAD_EGL_VERSION_1_1 VTK_GLAD_EGL_VERSION_1_1
#define GLAD_EGL_VERSION_1_2 VTK_GLAD_EGL_VERSION_1_2
#define GLAD_EGL_VERSION_1_3 VTK_GLAD_EGL_VERSION_1_3
#define GLAD_EGL_VERSION_1_4 VTK_GLAD_EGL_VERSION_1_4
#define GLAD_EGL_VERSION_1_5 VTK_GLAD_EGL_VERSION_1_5
#define GLAD_EGL_ANDROID_GLES_layers VTK_GLAD_EGL_ANDROID_GLES_layers
#define GLAD_EGL_ANDROID_blob_cache VTK_GLAD_EGL_ANDROID_blob_cache
#define GLAD_EGL_ANDROID_create_native_client_buffer                                               \
  VTK_GLAD_EGL_ANDROID_create_native_client_buffer
#define GLAD_EGL_ANDROID_framebuffer_target VTK_GLAD_EGL_ANDROID_framebuffer_target
#define GLAD_EGL_ANDROID_front_buffer_auto_refresh VTK_GLAD_EGL_ANDROID_front_buffer_auto_refresh
#define GLAD_EGL_ANDROID_get_frame_timestamps VTK_GLAD_EGL_ANDROID_get_frame_timestamps
#define GLAD_EGL_ANDROID_get_native_client_buffer VTK_GLAD_EGL_ANDROID_get_native_client_buffer
#define GLAD_EGL_ANDROID_image_native_buffer VTK_GLAD_EGL_ANDROID_image_native_buffer
#define GLAD_EGL_ANDROID_native_fence_sync VTK_GLAD_EGL_ANDROID_native_fence_sync
#define GLAD_EGL_ANDROID_presentation_time VTK_GLAD_EGL_ANDROID_presentation_time
#define GLAD_EGL_ANDROID_recordable VTK_GLAD_EGL_ANDROID_recordable
#define GLAD_EGL_ANDROID_telemetry_hint VTK_GLAD_EGL_ANDROID_telemetry_hint
#define GLAD_EGL_ANGLE_d3d_share_handle_client_buffer                                              \
  VTK_GLAD_EGL_ANGLE_d3d_share_handle_client_buffer
#define GLAD_EGL_ANGLE_device_d3d VTK_GLAD_EGL_ANGLE_device_d3d
#define GLAD_EGL_ANGLE_query_surface_pointer VTK_GLAD_EGL_ANGLE_query_surface_pointer
#define GLAD_EGL_ANGLE_surface_d3d_texture_2d_share_handle                                         \
  VTK_GLAD_EGL_ANGLE_surface_d3d_texture_2d_share_handle
#define GLAD_EGL_ANGLE_sync_control_rate VTK_GLAD_EGL_ANGLE_sync_control_rate
#define GLAD_EGL_ANGLE_window_fixed_size VTK_GLAD_EGL_ANGLE_window_fixed_size
#define GLAD_EGL_ARM_image_format VTK_GLAD_EGL_ARM_image_format
#define GLAD_EGL_ARM_implicit_external_sync VTK_GLAD_EGL_ARM_implicit_external_sync
#define GLAD_EGL_ARM_pixmap_multisample_discard VTK_GLAD_EGL_ARM_pixmap_multisample_discard
#define GLAD_EGL_EXT_bind_to_front VTK_GLAD_EGL_EXT_bind_to_front
#define GLAD_EGL_EXT_buffer_age VTK_GLAD_EGL_EXT_buffer_age
#define GLAD_EGL_EXT_client_extensions VTK_GLAD_EGL_EXT_client_extensions
#define GLAD_EGL_EXT_client_sync VTK_GLAD_EGL_EXT_client_sync
#define GLAD_EGL_EXT_compositor VTK_GLAD_EGL_EXT_compositor
#define GLAD_EGL_EXT_config_select_group VTK_GLAD_EGL_EXT_config_select_group
#define GLAD_EGL_EXT_create_context_robustness VTK_GLAD_EGL_EXT_create_context_robustness
#define GLAD_EGL_EXT_device_base VTK_GLAD_EGL_EXT_device_base
#define GLAD_EGL_EXT_device_drm VTK_GLAD_EGL_EXT_device_drm
#define GLAD_EGL_EXT_device_drm_render_node VTK_GLAD_EGL_EXT_device_drm_render_node
#define GLAD_EGL_EXT_device_enumeration VTK_GLAD_EGL_EXT_device_enumeration
#define GLAD_EGL_EXT_device_openwf VTK_GLAD_EGL_EXT_device_openwf
#define GLAD_EGL_EXT_device_persistent_id VTK_GLAD_EGL_EXT_device_persistent_id
#define GLAD_EGL_EXT_device_query VTK_GLAD_EGL_EXT_device_query
#define GLAD_EGL_EXT_device_query_name VTK_GLAD_EGL_EXT_device_query_name
#define GLAD_EGL_EXT_explicit_device VTK_GLAD_EGL_EXT_explicit_device
#define GLAD_EGL_EXT_gl_colorspace_bt2020_hlg VTK_GLAD_EGL_EXT_gl_colorspace_bt2020_hlg
#define GLAD_EGL_EXT_gl_colorspace_bt2020_linear VTK_GLAD_EGL_EXT_gl_colorspace_bt2020_linear
#define GLAD_EGL_EXT_gl_colorspace_bt2020_pq VTK_GLAD_EGL_EXT_gl_colorspace_bt2020_pq
#define GLAD_EGL_EXT_gl_colorspace_display_p3 VTK_GLAD_EGL_EXT_gl_colorspace_display_p3
#define GLAD_EGL_EXT_gl_colorspace_display_p3_linear                                               \
  VTK_GLAD_EGL_EXT_gl_colorspace_display_p3_linear
#define GLAD_EGL_EXT_gl_colorspace_display_p3_passthrough                                          \
  VTK_GLAD_EGL_EXT_gl_colorspace_display_p3_passthrough
#define GLAD_EGL_EXT_gl_colorspace_scrgb VTK_GLAD_EGL_EXT_gl_colorspace_scrgb
#define GLAD_EGL_EXT_gl_colorspace_scrgb_linear VTK_GLAD_EGL_EXT_gl_colorspace_scrgb_linear
#define GLAD_EGL_EXT_image_dma_buf_import VTK_GLAD_EGL_EXT_image_dma_buf_import
#define GLAD_EGL_EXT_image_dma_buf_import_modifiers VTK_GLAD_EGL_EXT_image_dma_buf_import_modifiers
#define GLAD_EGL_EXT_image_gl_colorspace VTK_GLAD_EGL_EXT_image_gl_colorspace
#define GLAD_EGL_EXT_image_implicit_sync_control VTK_GLAD_EGL_EXT_image_implicit_sync_control
#define GLAD_EGL_EXT_multiview_window VTK_GLAD_EGL_EXT_multiview_window
#define GLAD_EGL_EXT_output_base VTK_GLAD_EGL_EXT_output_base
#define GLAD_EGL_EXT_output_drm VTK_GLAD_EGL_EXT_output_drm
#define GLAD_EGL_EXT_output_openwf VTK_GLAD_EGL_EXT_output_openwf
#define GLAD_EGL_EXT_pixel_format_float VTK_GLAD_EGL_EXT_pixel_format_float
#define GLAD_EGL_EXT_platform_base VTK_GLAD_EGL_EXT_platform_base
#define GLAD_EGL_EXT_platform_device VTK_GLAD_EGL_EXT_platform_device
#define GLAD_EGL_EXT_platform_wayland VTK_GLAD_EGL_EXT_platform_wayland
#define GLAD_EGL_EXT_platform_x11 VTK_GLAD_EGL_EXT_platform_x11
#define GLAD_EGL_EXT_platform_xcb VTK_GLAD_EGL_EXT_platform_xcb
#define GLAD_EGL_EXT_present_opaque VTK_GLAD_EGL_EXT_present_opaque
#define GLAD_EGL_EXT_protected_content VTK_GLAD_EGL_EXT_protected_content
#define GLAD_EGL_EXT_protected_surface VTK_GLAD_EGL_EXT_protected_surface
#define GLAD_EGL_EXT_query_reset_notification_strategy                                             \
  VTK_GLAD_EGL_EXT_query_reset_notification_strategy
#define GLAD_EGL_EXT_stream_consumer_egloutput VTK_GLAD_EGL_EXT_stream_consumer_egloutput
#define GLAD_EGL_EXT_surface_CTA861_3_metadata VTK_GLAD_EGL_EXT_surface_CTA861_3_metadata
#define GLAD_EGL_EXT_surface_SMPTE2086_metadata VTK_GLAD_EGL_EXT_surface_SMPTE2086_metadata
#define GLAD_EGL_EXT_surface_compression VTK_GLAD_EGL_EXT_surface_compression
#define GLAD_EGL_EXT_swap_buffers_with_damage VTK_GLAD_EGL_EXT_swap_buffers_with_damage
#define GLAD_EGL_EXT_sync_reuse VTK_GLAD_EGL_EXT_sync_reuse
#define GLAD_EGL_EXT_yuv_surface VTK_GLAD_EGL_EXT_yuv_surface
#define GLAD_EGL_HI_clientpixmap VTK_GLAD_EGL_HI_clientpixmap
#define GLAD_EGL_HI_colorformats VTK_GLAD_EGL_HI_colorformats
#define GLAD_EGL_IMG_context_priority VTK_GLAD_EGL_IMG_context_priority
#define GLAD_EGL_IMG_image_plane_attribs VTK_GLAD_EGL_IMG_image_plane_attribs
#define GLAD_EGL_KHR_cl_event VTK_GLAD_EGL_KHR_cl_event
#define GLAD_EGL_KHR_cl_event2 VTK_GLAD_EGL_KHR_cl_event2
#define GLAD_EGL_KHR_client_get_all_proc_addresses VTK_GLAD_EGL_KHR_client_get_all_proc_addresses
#define GLAD_EGL_KHR_config_attribs VTK_GLAD_EGL_KHR_config_attribs
#define GLAD_EGL_KHR_context_flush_control VTK_GLAD_EGL_KHR_context_flush_control
#define GLAD_EGL_KHR_create_context VTK_GLAD_EGL_KHR_create_context
#define GLAD_EGL_KHR_create_context_no_error VTK_GLAD_EGL_KHR_create_context_no_error
#define GLAD_EGL_KHR_debug VTK_GLAD_EGL_KHR_debug
#define GLAD_EGL_KHR_display_reference VTK_GLAD_EGL_KHR_display_reference
#define GLAD_EGL_KHR_fence_sync VTK_GLAD_EGL_KHR_fence_sync
#define GLAD_EGL_KHR_get_all_proc_addresses VTK_GLAD_EGL_KHR_get_all_proc_addresses
#define GLAD_EGL_KHR_gl_colorspace VTK_GLAD_EGL_KHR_gl_colorspace
#define GLAD_EGL_KHR_gl_renderbuffer_image VTK_GLAD_EGL_KHR_gl_renderbuffer_image
#define GLAD_EGL_KHR_gl_texture_2D_image VTK_GLAD_EGL_KHR_gl_texture_2D_image
#define GLAD_EGL_KHR_gl_texture_3D_image VTK_GLAD_EGL_KHR_gl_texture_3D_image
#define GLAD_EGL_KHR_gl_texture_cubemap_image VTK_GLAD_EGL_KHR_gl_texture_cubemap_image
#define GLAD_EGL_KHR_image VTK_GLAD_EGL_KHR_image
#define GLAD_EGL_KHR_image_base VTK_GLAD_EGL_KHR_image_base
#define GLAD_EGL_KHR_image_pixmap VTK_GLAD_EGL_KHR_image_pixmap
#define GLAD_EGL_KHR_lock_surface VTK_GLAD_EGL_KHR_lock_surface
#define GLAD_EGL_KHR_lock_surface2 VTK_GLAD_EGL_KHR_lock_surface2
#define GLAD_EGL_KHR_lock_surface3 VTK_GLAD_EGL_KHR_lock_surface3
#define GLAD_EGL_KHR_mutable_render_buffer VTK_GLAD_EGL_KHR_mutable_render_buffer
#define GLAD_EGL_KHR_no_config_context VTK_GLAD_EGL_KHR_no_config_context
#define GLAD_EGL_KHR_partial_update VTK_GLAD_EGL_KHR_partial_update
#define GLAD_EGL_KHR_platform_android VTK_GLAD_EGL_KHR_platform_android
#define GLAD_EGL_KHR_platform_gbm VTK_GLAD_EGL_KHR_platform_gbm
#define GLAD_EGL_KHR_platform_wayland VTK_GLAD_EGL_KHR_platform_wayland
#define GLAD_EGL_KHR_platform_x11 VTK_GLAD_EGL_KHR_platform_x11
#define GLAD_EGL_KHR_reusable_sync VTK_GLAD_EGL_KHR_reusable_sync
#define GLAD_EGL_KHR_stream VTK_GLAD_EGL_KHR_stream
#define GLAD_EGL_KHR_stream_attrib VTK_GLAD_EGL_KHR_stream_attrib
#define GLAD_EGL_KHR_stream_consumer_gltexture VTK_GLAD_EGL_KHR_stream_consumer_gltexture
#define GLAD_EGL_KHR_stream_cross_process_fd VTK_GLAD_EGL_KHR_stream_cross_process_fd
#define GLAD_EGL_KHR_stream_fifo VTK_GLAD_EGL_KHR_stream_fifo
#define GLAD_EGL_KHR_stream_producer_aldatalocator VTK_GLAD_EGL_KHR_stream_producer_aldatalocator
#define GLAD_EGL_KHR_stream_producer_eglsurface VTK_GLAD_EGL_KHR_stream_producer_eglsurface
#define GLAD_EGL_KHR_surfaceless_context VTK_GLAD_EGL_KHR_surfaceless_context
#define GLAD_EGL_KHR_swap_buffers_with_damage VTK_GLAD_EGL_KHR_swap_buffers_with_damage
#define GLAD_EGL_KHR_vg_parent_image VTK_GLAD_EGL_KHR_vg_parent_image
#define GLAD_EGL_KHR_wait_sync VTK_GLAD_EGL_KHR_wait_sync
#define GLAD_EGL_MESA_drm_image VTK_GLAD_EGL_MESA_drm_image
#define GLAD_EGL_MESA_image_dma_buf_export VTK_GLAD_EGL_MESA_image_dma_buf_export
#define GLAD_EGL_MESA_platform_gbm VTK_GLAD_EGL_MESA_platform_gbm
#define GLAD_EGL_MESA_platform_surfaceless VTK_GLAD_EGL_MESA_platform_surfaceless
#define GLAD_EGL_MESA_query_driver VTK_GLAD_EGL_MESA_query_driver
#define GLAD_EGL_NOK_swap_region VTK_GLAD_EGL_NOK_swap_region
#define GLAD_EGL_NOK_swap_region2 VTK_GLAD_EGL_NOK_swap_region2
#define GLAD_EGL_NOK_texture_from_pixmap VTK_GLAD_EGL_NOK_texture_from_pixmap
#define GLAD_EGL_NV_3dvision_surface VTK_GLAD_EGL_NV_3dvision_surface
#define GLAD_EGL_NV_context_priority_realtime VTK_GLAD_EGL_NV_context_priority_realtime
#define GLAD_EGL_NV_coverage_sample VTK_GLAD_EGL_NV_coverage_sample
#define GLAD_EGL_NV_coverage_sample_resolve VTK_GLAD_EGL_NV_coverage_sample_resolve
#define GLAD_EGL_NV_cuda_event VTK_GLAD_EGL_NV_cuda_event
#define GLAD_EGL_NV_depth_nonlinear VTK_GLAD_EGL_NV_depth_nonlinear
#define GLAD_EGL_NV_device_cuda VTK_GLAD_EGL_NV_device_cuda
#define GLAD_EGL_NV_native_query VTK_GLAD_EGL_NV_native_query
#define GLAD_EGL_NV_post_convert_rounding VTK_GLAD_EGL_NV_post_convert_rounding
#define GLAD_EGL_NV_post_sub_buffer VTK_GLAD_EGL_NV_post_sub_buffer
#define GLAD_EGL_NV_quadruple_buffer VTK_GLAD_EGL_NV_quadruple_buffer
#define GLAD_EGL_NV_robustness_video_memory_purge VTK_GLAD_EGL_NV_robustness_video_memory_purge
#define GLAD_EGL_NV_stream_consumer_eglimage VTK_GLAD_EGL_NV_stream_consumer_eglimage
#define GLAD_EGL_NV_stream_consumer_eglimage_use_scanout_attrib                                    \
  VTK_GLAD_EGL_NV_stream_consumer_eglimage_use_scanout_attrib
#define GLAD_EGL_NV_stream_consumer_gltexture_yuv VTK_GLAD_EGL_NV_stream_consumer_gltexture_yuv
#define GLAD_EGL_NV_stream_cross_display VTK_GLAD_EGL_NV_stream_cross_display
#define GLAD_EGL_NV_stream_cross_object VTK_GLAD_EGL_NV_stream_cross_object
#define GLAD_EGL_NV_stream_cross_partition VTK_GLAD_EGL_NV_stream_cross_partition
#define GLAD_EGL_NV_stream_cross_process VTK_GLAD_EGL_NV_stream_cross_process
#define GLAD_EGL_NV_stream_cross_system VTK_GLAD_EGL_NV_stream_cross_system
#define GLAD_EGL_NV_stream_dma VTK_GLAD_EGL_NV_stream_dma
#define GLAD_EGL_NV_stream_fifo_next VTK_GLAD_EGL_NV_stream_fifo_next
#define GLAD_EGL_NV_stream_fifo_synchronous VTK_GLAD_EGL_NV_stream_fifo_synchronous
#define GLAD_EGL_NV_stream_flush VTK_GLAD_EGL_NV_stream_flush
#define GLAD_EGL_NV_stream_frame_limits VTK_GLAD_EGL_NV_stream_frame_limits
#define GLAD_EGL_NV_stream_metadata VTK_GLAD_EGL_NV_stream_metadata
#define GLAD_EGL_NV_stream_origin VTK_GLAD_EGL_NV_stream_origin
#define GLAD_EGL_NV_stream_remote VTK_GLAD_EGL_NV_stream_remote
#define GLAD_EGL_NV_stream_reset VTK_GLAD_EGL_NV_stream_reset
#define GLAD_EGL_NV_stream_socket VTK_GLAD_EGL_NV_stream_socket
#define GLAD_EGL_NV_stream_socket_inet VTK_GLAD_EGL_NV_stream_socket_inet
#define GLAD_EGL_NV_stream_socket_unix VTK_GLAD_EGL_NV_stream_socket_unix
#define GLAD_EGL_NV_stream_sync VTK_GLAD_EGL_NV_stream_sync
#define GLAD_EGL_NV_sync VTK_GLAD_EGL_NV_sync
#define GLAD_EGL_NV_system_time VTK_GLAD_EGL_NV_system_time
#define GLAD_EGL_NV_triple_buffer VTK_GLAD_EGL_NV_triple_buffer
#define GLAD_EGL_QNX_image_native_buffer VTK_GLAD_EGL_QNX_image_native_buffer
#define GLAD_EGL_QNX_platform_screen VTK_GLAD_EGL_QNX_platform_screen
#define GLAD_EGL_TIZEN_image_native_buffer VTK_GLAD_EGL_TIZEN_image_native_buffer
#define GLAD_EGL_TIZEN_image_native_surface VTK_GLAD_EGL_TIZEN_image_native_surface
#define GLAD_EGL_WL_bind_wayland_display VTK_GLAD_EGL_WL_bind_wayland_display
#define GLAD_EGL_WL_create_wayland_buffer_from_image                                               \
  VTK_GLAD_EGL_WL_create_wayland_buffer_from_image

#define glad_eglBindAPI vtk_glad_eglBindAPI
#define glad_eglBindTexImage vtk_glad_eglBindTexImage
#define glad_eglBindWaylandDisplayWL vtk_glad_eglBindWaylandDisplayWL
#define glad_eglChooseConfig vtk_glad_eglChooseConfig
#define glad_eglClientSignalSyncEXT vtk_glad_eglClientSignalSyncEXT
#define glad_eglClientWaitSync vtk_glad_eglClientWaitSync
#define glad_eglClientWaitSyncKHR vtk_glad_eglClientWaitSyncKHR
#define glad_eglClientWaitSyncNV vtk_glad_eglClientWaitSyncNV
#define glad_eglCompositorBindTexWindowEXT vtk_glad_eglCompositorBindTexWindowEXT
#define glad_eglCompositorSetContextAttributesEXT vtk_glad_eglCompositorSetContextAttributesEXT
#define glad_eglCompositorSetContextListEXT vtk_glad_eglCompositorSetContextListEXT
#define glad_eglCompositorSetSizeEXT vtk_glad_eglCompositorSetSizeEXT
#define glad_eglCompositorSetWindowAttributesEXT vtk_glad_eglCompositorSetWindowAttributesEXT
#define glad_eglCompositorSetWindowListEXT vtk_glad_eglCompositorSetWindowListEXT
#define glad_eglCompositorSwapPolicyEXT vtk_glad_eglCompositorSwapPolicyEXT
#define glad_eglCopyBuffers vtk_glad_eglCopyBuffers
#define glad_eglCreateContext vtk_glad_eglCreateContext
#define glad_eglCreateDRMImageMESA vtk_glad_eglCreateDRMImageMESA
#define glad_eglCreateFenceSyncNV vtk_glad_eglCreateFenceSyncNV
#define glad_eglCreateImage vtk_glad_eglCreateImage
#define glad_eglCreateImageKHR vtk_glad_eglCreateImageKHR
#define glad_eglCreateNativeClientBufferANDROID vtk_glad_eglCreateNativeClientBufferANDROID
#define glad_eglCreatePbufferFromClientBuffer vtk_glad_eglCreatePbufferFromClientBuffer
#define glad_eglCreatePbufferSurface vtk_glad_eglCreatePbufferSurface
#define glad_eglCreatePixmapSurface vtk_glad_eglCreatePixmapSurface
#define glad_eglCreatePixmapSurfaceHI vtk_glad_eglCreatePixmapSurfaceHI
#define glad_eglCreatePlatformPixmapSurface vtk_glad_eglCreatePlatformPixmapSurface
#define glad_eglCreatePlatformPixmapSurfaceEXT vtk_glad_eglCreatePlatformPixmapSurfaceEXT
#define glad_eglCreatePlatformWindowSurface vtk_glad_eglCreatePlatformWindowSurface
#define glad_eglCreatePlatformWindowSurfaceEXT vtk_glad_eglCreatePlatformWindowSurfaceEXT
#define glad_eglCreateStreamAttribKHR vtk_glad_eglCreateStreamAttribKHR
#define glad_eglCreateStreamFromFileDescriptorKHR vtk_glad_eglCreateStreamFromFileDescriptorKHR
#define glad_eglCreateStreamKHR vtk_glad_eglCreateStreamKHR
#define glad_eglCreateStreamProducerSurfaceKHR vtk_glad_eglCreateStreamProducerSurfaceKHR
#define glad_eglCreateStreamSyncNV vtk_glad_eglCreateStreamSyncNV
#define glad_eglCreateSync vtk_glad_eglCreateSync
#define glad_eglCreateSync64KHR vtk_glad_eglCreateSync64KHR
#define glad_eglCreateSyncKHR vtk_glad_eglCreateSyncKHR
#define glad_eglCreateWaylandBufferFromImageWL vtk_glad_eglCreateWaylandBufferFromImageWL
#define glad_eglCreateWindowSurface vtk_glad_eglCreateWindowSurface
#define glad_eglDebugMessageControlKHR vtk_glad_eglDebugMessageControlKHR
#define glad_eglDestroyContext vtk_glad_eglDestroyContext
#define glad_eglDestroyImage vtk_glad_eglDestroyImage
#define glad_eglDestroyImageKHR vtk_glad_eglDestroyImageKHR
#define glad_eglDestroyStreamKHR vtk_glad_eglDestroyStreamKHR
#define glad_eglDestroySurface vtk_glad_eglDestroySurface
#define glad_eglDestroySync vtk_glad_eglDestroySync
#define glad_eglDestroySyncKHR vtk_glad_eglDestroySyncKHR
#define glad_eglDestroySyncNV vtk_glad_eglDestroySyncNV
#define glad_eglDupNativeFenceFDANDROID vtk_glad_eglDupNativeFenceFDANDROID
#define glad_eglExportDMABUFImageMESA vtk_glad_eglExportDMABUFImageMESA
#define glad_eglExportDMABUFImageQueryMESA vtk_glad_eglExportDMABUFImageQueryMESA
#define glad_eglExportDRMImageMESA vtk_glad_eglExportDRMImageMESA
#define glad_eglFenceNV vtk_glad_eglFenceNV
#define glad_eglGetCompositorTimingANDROID vtk_glad_eglGetCompositorTimingANDROID
#define glad_eglGetCompositorTimingSupportedANDROID vtk_glad_eglGetCompositorTimingSupportedANDROID
#define glad_eglGetConfigAttrib vtk_glad_eglGetConfigAttrib
#define glad_eglGetConfigs vtk_glad_eglGetConfigs
#define glad_eglGetCurrentContext vtk_glad_eglGetCurrentContext
#define glad_eglGetCurrentDisplay vtk_glad_eglGetCurrentDisplay
#define glad_eglGetCurrentSurface vtk_glad_eglGetCurrentSurface
#define glad_eglGetDisplay vtk_glad_eglGetDisplay
#define glad_eglGetDisplayDriverConfig vtk_glad_eglGetDisplayDriverConfig
#define glad_eglGetDisplayDriverName vtk_glad_eglGetDisplayDriverName
#define glad_eglGetError vtk_glad_eglGetError
#define glad_eglGetFrameTimestampSupportedANDROID vtk_glad_eglGetFrameTimestampSupportedANDROID
#define glad_eglGetFrameTimestampsANDROID vtk_glad_eglGetFrameTimestampsANDROID
#define glad_eglGetMscRateANGLE vtk_glad_eglGetMscRateANGLE
#define glad_eglGetNativeClientBufferANDROID vtk_glad_eglGetNativeClientBufferANDROID
#define glad_eglGetNextFrameIdANDROID vtk_glad_eglGetNextFrameIdANDROID
#define glad_eglGetOutputLayersEXT vtk_glad_eglGetOutputLayersEXT
#define glad_eglGetOutputPortsEXT vtk_glad_eglGetOutputPortsEXT
#define glad_eglGetPlatformDisplay vtk_glad_eglGetPlatformDisplay
#define glad_eglGetPlatformDisplayEXT vtk_glad_eglGetPlatformDisplayEXT
#define glad_eglGetProcAddress vtk_glad_eglGetProcAddress
#define glad_eglGetStreamFileDescriptorKHR vtk_glad_eglGetStreamFileDescriptorKHR
#define glad_eglGetSyncAttrib vtk_glad_eglGetSyncAttrib
#define glad_eglGetSyncAttribKHR vtk_glad_eglGetSyncAttribKHR
#define glad_eglGetSyncAttribNV vtk_glad_eglGetSyncAttribNV
#define glad_eglGetSystemTimeFrequencyNV vtk_glad_eglGetSystemTimeFrequencyNV
#define glad_eglGetSystemTimeNV vtk_glad_eglGetSystemTimeNV
#define glad_eglInitialize vtk_glad_eglInitialize
#define glad_eglLabelObjectKHR vtk_glad_eglLabelObjectKHR
#define glad_eglLockSurfaceKHR vtk_glad_eglLockSurfaceKHR
#define glad_eglMakeCurrent vtk_glad_eglMakeCurrent
#define glad_eglOutputLayerAttribEXT vtk_glad_eglOutputLayerAttribEXT
#define glad_eglOutputPortAttribEXT vtk_glad_eglOutputPortAttribEXT
#define glad_eglPostSubBufferNV vtk_glad_eglPostSubBufferNV
#define glad_eglPresentationTimeANDROID vtk_glad_eglPresentationTimeANDROID
#define glad_eglQueryAPI vtk_glad_eglQueryAPI
#define glad_eglQueryContext vtk_glad_eglQueryContext
#define glad_eglQueryDebugKHR vtk_glad_eglQueryDebugKHR
#define glad_eglQueryDeviceAttribEXT vtk_glad_eglQueryDeviceAttribEXT
#define glad_eglQueryDeviceBinaryEXT vtk_glad_eglQueryDeviceBinaryEXT
#define glad_eglQueryDeviceStringEXT vtk_glad_eglQueryDeviceStringEXT
#define glad_eglQueryDevicesEXT vtk_glad_eglQueryDevicesEXT
#define glad_eglQueryDisplayAttribEXT vtk_glad_eglQueryDisplayAttribEXT
#define glad_eglQueryDisplayAttribKHR vtk_glad_eglQueryDisplayAttribKHR
#define glad_eglQueryDisplayAttribNV vtk_glad_eglQueryDisplayAttribNV
#define glad_eglQueryDmaBufFormatsEXT vtk_glad_eglQueryDmaBufFormatsEXT
#define glad_eglQueryDmaBufModifiersEXT vtk_glad_eglQueryDmaBufModifiersEXT
#define glad_eglQueryNativeDisplayNV vtk_glad_eglQueryNativeDisplayNV
#define glad_eglQueryNativePixmapNV vtk_glad_eglQueryNativePixmapNV
#define glad_eglQueryNativeWindowNV vtk_glad_eglQueryNativeWindowNV
#define glad_eglQueryOutputLayerAttribEXT vtk_glad_eglQueryOutputLayerAttribEXT
#define glad_eglQueryOutputLayerStringEXT vtk_glad_eglQueryOutputLayerStringEXT
#define glad_eglQueryOutputPortAttribEXT vtk_glad_eglQueryOutputPortAttribEXT
#define glad_eglQueryOutputPortStringEXT vtk_glad_eglQueryOutputPortStringEXT
#define glad_eglQueryStreamAttribKHR vtk_glad_eglQueryStreamAttribKHR
#define glad_eglQueryStreamConsumerEventNV vtk_glad_eglQueryStreamConsumerEventNV
#define glad_eglQueryStreamKHR vtk_glad_eglQueryStreamKHR
#define glad_eglQueryStreamMetadataNV vtk_glad_eglQueryStreamMetadataNV
#define glad_eglQueryStreamTimeKHR vtk_glad_eglQueryStreamTimeKHR
#define glad_eglQueryStreamu64KHR vtk_glad_eglQueryStreamu64KHR
#define glad_eglQueryString vtk_glad_eglQueryString
#define glad_eglQuerySupportedCompressionRatesEXT vtk_glad_eglQuerySupportedCompressionRatesEXT
#define glad_eglQuerySurface vtk_glad_eglQuerySurface
#define glad_eglQuerySurface64KHR vtk_glad_eglQuerySurface64KHR
#define glad_eglQuerySurfacePointerANGLE vtk_glad_eglQuerySurfacePointerANGLE
#define glad_eglQueryWaylandBufferWL vtk_glad_eglQueryWaylandBufferWL
#define glad_eglReleaseTexImage vtk_glad_eglReleaseTexImage
#define glad_eglReleaseThread vtk_glad_eglReleaseThread
#define glad_eglResetStreamNV vtk_glad_eglResetStreamNV
#define glad_eglSetBlobCacheFuncsANDROID vtk_glad_eglSetBlobCacheFuncsANDROID
#define glad_eglSetDamageRegionKHR vtk_glad_eglSetDamageRegionKHR
#define glad_eglSetStreamAttribKHR vtk_glad_eglSetStreamAttribKHR
#define glad_eglSetStreamMetadataNV vtk_glad_eglSetStreamMetadataNV
#define glad_eglSignalSyncKHR vtk_glad_eglSignalSyncKHR
#define glad_eglSignalSyncNV vtk_glad_eglSignalSyncNV
#define glad_eglStreamAcquireImageNV vtk_glad_eglStreamAcquireImageNV
#define glad_eglStreamAttribKHR vtk_glad_eglStreamAttribKHR
#define glad_eglStreamConsumerAcquireAttribKHR vtk_glad_eglStreamConsumerAcquireAttribKHR
#define glad_eglStreamConsumerAcquireKHR vtk_glad_eglStreamConsumerAcquireKHR
#define glad_eglStreamConsumerGLTextureExternalAttribsNV                                           \
  vtk_glad_eglStreamConsumerGLTextureExternalAttribsNV
#define glad_eglStreamConsumerGLTextureExternalKHR vtk_glad_eglStreamConsumerGLTextureExternalKHR
#define glad_eglStreamConsumerOutputEXT vtk_glad_eglStreamConsumerOutputEXT
#define glad_eglStreamConsumerReleaseAttribKHR vtk_glad_eglStreamConsumerReleaseAttribKHR
#define glad_eglStreamConsumerReleaseKHR vtk_glad_eglStreamConsumerReleaseKHR
#define glad_eglStreamFlushNV vtk_glad_eglStreamFlushNV
#define glad_eglStreamImageConsumerConnectNV vtk_glad_eglStreamImageConsumerConnectNV
#define glad_eglStreamReleaseImageNV vtk_glad_eglStreamReleaseImageNV
#define glad_eglSurfaceAttrib vtk_glad_eglSurfaceAttrib
#define glad_eglSwapBuffers vtk_glad_eglSwapBuffers
#define glad_eglSwapBuffersRegion2NOK vtk_glad_eglSwapBuffersRegion2NOK
#define glad_eglSwapBuffersRegionNOK vtk_glad_eglSwapBuffersRegionNOK
#define glad_eglSwapBuffersWithDamageEXT vtk_glad_eglSwapBuffersWithDamageEXT
#define glad_eglSwapBuffersWithDamageKHR vtk_glad_eglSwapBuffersWithDamageKHR
#define glad_eglSwapInterval vtk_glad_eglSwapInterval
#define glad_eglTerminate vtk_glad_eglTerminate
#define glad_eglUnbindWaylandDisplayWL vtk_glad_eglUnbindWaylandDisplayWL
#define glad_eglUnlockSurfaceKHR vtk_glad_eglUnlockSurfaceKHR
#define glad_eglUnsignalSyncEXT vtk_glad_eglUnsignalSyncEXT
#define glad_eglWaitClient vtk_glad_eglWaitClient
#define glad_eglWaitGL vtk_glad_eglWaitGL
#define glad_eglWaitNative vtk_glad_eglWaitNative
#define glad_eglWaitSync vtk_glad_eglWaitSync
#define glad_eglWaitSyncKHR vtk_glad_eglWaitSyncKHR

#define gladLoadEGLUserPtr vtk_gladLoadEGLUserPtr
#define gladLoadEGL vtk_gladLoadEGL
#define gladLoaderLoadEGL vtk_gladLoaderLoadEGL
#define gladLoaderUnloadEGL vtk_gladLoaderUnloadEGL

#endif
